import simplejson
from Acquisition import aq_inner
from zope.interface import implements
from zope.component import getMultiAdapter, getUtility
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from plone.memoize import view
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import IPloneSiteRoot
from collective.virtualtreecategories.browser.interfaces import IVirtualTreeCategoriesSettingsView
from collective.virtualtreecategories.interfaces import IVirtualTreeCategoryConfiguration
from collective.virtualtreecategories.config import CATEGORY_SPLITTER
import logging
from sets import Set

logger = logging.getLogger('vtc-controlpanel')
        
class VirtualTreeCategoriesSettingsView(BrowserView):
    implements(IVirtualTreeCategoriesSettingsView)

    template=ViewPageTemplateFile('controlpanel.pt')

    @view.memoize_contextless
    def tools(self):
        """ returns tools view. Available items are all portal_xxxxx 
            For example: catalog, membership, url
            http://dev.plone.org/plone/browser/plone.app.layout/trunk/plone/app/layout/globals/tools.py
        """
        return getMultiAdapter((self.context, self.request), name=u"plone_tools")

    @view.memoize_contextless
    def portal_state(self):
        """ returns 
            http://dev.plone.org/plone/browser/plone.app.layout/trunk/plone/app/layout/globals/portal.py
        """
        return getMultiAdapter((self.context, self.request), name=u"plone_portal_state")

    def all_keywords(self):
        vals =  list(self.tools().catalog().uniqueValuesFor('Subject'))
        vals.sort()
        return vals
        
class CategoryKeywords(BrowserView):
    
    @view.memoize_contextless
    def tools(self):
        """ returns tools view. Available items are all portal_xxxxx 
            For example: catalog, membership, url
            http://dev.plone.org/plone/browser/plone.app.layout/trunk/plone/app/layout/globals/tools.py
        """
        return getMultiAdapter((self.context, self.request), name=u"plone_tools")

    def _category_path_from_request(self, path=None):
        if path is None:
            category_path = self.request.form.get('category_path')
        else:
            category_path = path
        if not category_path:
            return []
        if isinstance(category_path, basestring):
            category_path = category_path.split(',')
        category_path.reverse()
        # omit root node
        category_path = category_path and category_path[1:]
        return category_path
    
    def get_category_keywords(self):
        self.request.response.setHeader('Content-Type', 'application/json; charset=utf-8')
        portal = getUtility(IPloneSiteRoot)
        category_path = self._category_path_from_request()
        if not category_path:
            kws = []
        else:
            kws = IVirtualTreeCategoryConfiguration(portal).get(category_path)
        return simplejson.dumps(dict(keywords = kws))

    def save_category_keywords(self):
        portal = getUtility(IPloneSiteRoot)
        category_path = self._category_path_from_request()
        kws = self.request.form.get('kws')
        if isinstance(kws, basestring):
            kws = kws.split(',')
        if not category_path:
            return 'No valid category selected'
        logger.info('Going to save %d keywords to category %r' % (len(kws), category_path))
        IVirtualTreeCategoryConfiguration(portal).set(category_path, kws)
        self.request.response.setHeader('Content-Type', 'text/plain; charset=utf-8')
        return 'Category %s saved' % category_path[-1]

    def categories_tree(self):
        self.request.response.setHeader('Content-Type', 'application/json; charset=utf-8')
        storage = IVirtualTreeCategoryConfiguration(getUtility(IPloneSiteRoot))
        # inject root node
        root = dict(
          attributes = { 'id' : "root-node", 'rel': 'root' }, 
          state = "open", 
          data = "Root node",
          children = storage.category_tree()
         )
        return simplejson.dumps(root)

    def category_added_renamed(self):
        storage = IVirtualTreeCategoryConfiguration(getUtility(IPloneSiteRoot))
        old_id = self.request.form.get('old_id')
        new_name = self.request.form.get('new_name')
        if not new_name:
            return 'Missing new name.'
        if not old_id:
            # create new node
            # strip last element from category_path, because it is newly created category name which does not exist yet
            category_path = self._category_path_from_request()[:-1]
            new_id = storage.add_category(category_path, new_name)
            if new_id:
                result = simplejson.dumps(dict(
                                           msg = 'Category created',
                                           new_id = new_id,
                                           result = True))
            else:
                result = simplejson.dumps(dict(
                                           msg = 'Category creation error!',
                                           new_id = '',
                                           result = False))
        else:
            # rename old node
            category_path = self._category_path_from_request()
            new_id = storage.rename_category(category_path, old_id, new_name)
            if new_id:
                result = simplejson.dumps(dict(
                                               msg = 'Category renamed',
                                               new_id = new_id,
                                               result = True))
            else:
                result = simplejson.dumps(dict(
                                               msg = 'Could not rename category.',
                                               new_id = '',
                                               result = False))
        return result
        
    def category_removed(self):
        storage = IVirtualTreeCategoryConfiguration(getUtility(IPloneSiteRoot))
        category_path = self._category_path_from_request()
        if not category_path:
           msg = 'You can''t not remove root category. Please reload page.'
        else:
            result = storage.remove_category(category_path)
            if result:
               msg = 'Category removed.'
            else:
               msg = 'Could not remove category. Please reload page.'
        return simplejson.dumps(dict( msg = msg,
                                      result = result))

    def list_keywords_by_categories(self):
        storage = IVirtualTreeCategoryConfiguration(getUtility(IPloneSiteRoot))
        # categories - list of lists where the sublist is category_path
        categories = self.request.form.get('categories', [])
        # list of keywords already assigned to the content
        selected   = self.request.form.get('selected', [])
        if isinstance(categories, basestring):
            categories = [categories]
        if isinstance(selected, basestring):
            selected = [selected]
        result = set()
        if categories:
            for category in categories:
                path = self._category_path_from_request(category)
                result.update(storage.get(path))
        else:
            result =  set(self.tools().catalog().uniqueValuesFor('Subject'))
            sorted(result)
        result = result.difference(selected)
        return simplejson.dumps(dict(keywords=list(result)))



from zope.i18n import MessageFactory
VTCMessageFactory = MessageFactory('collective.virtualtreecategories')

from collective.virtualtreecategories import widget

def initialize(context):
    """Initializer called when used as a Zope 2 product."""



from zope import schema
from zope.interface import Interface
from plone.theme.interfaces import IDefaultPloneLayer

class IVirtualTreeCategoriesSpecific(IDefaultPloneLayer):
    """Marker interface that defines a Zope 3 browser layer.
    """

class IVirtualTreeCategoryConfiguration(Interface):
    """ Configuration adapter which allows to set/read categories 
        from the stroage """
        
    def set(category_path, keywords):
        """ Set (assing) keywords to the category. 
            @category_path: string in form Level1/Level11/Level111 or
                            list ['Level1', 'Level11', 'Level111']
            @keywords: List of keywords to be assigned to this category
            Returns True/False
            """
            
    def get(category_path):
        """ Returns list of keywords assigned to the category.
            @category_path: string in form Level1/Level11/Level111 or
                            list ['Level1', 'Level11', 'Level111']
            """
    
    def category_tree():
        """ Returns categories in the tree form suitable for serializing using XML or JSON.
            This is JSON format accepted by the jsTree component:
            http://www.jstree.com/reference/_examples/1_datasources.html
                [
                 {
                  attributes: { id : "category1", [attribute : "attribute_value"] }, 
                  state: "closed" or "open", 
                  data: "Category 1"
                 },
                 {
                  attributes: { id : "category2", [attribute : "attribute_value"] }, 
                  state: "closed" or "open", 
                  data: "Category 2",
                  children: [
                                 {
                                  attributes: { id : "category2a", [attribute : "attribute_value"] }, 
                                  state: "closed" or "open", 
                                  data: "Category 2-A"
                                 }
                            ]
                 },
               ]

            Implementation note: There must be exactly one root node in the
            default (empty) jsTree to be able to create new nodes from the root
            using right-click. The root node should be added by controlpanel
            component, because it is not stored in the storage and it is jsTree
            requirement. Another JS component may allow creation of nodes
            without root category or it is possible to show button "Create new
            node". Another option is to store root level node and generate tree
            including this node..
        """

    def add_category(category_path, category_name):
        """ Create new category node as child of the category_path
            Returns new category id.
        """

    def remove_category(category_path): 
        """ Remove node at the end of category path including all child
            nodes 
            Returns True/False
        """

    def rename_category(category_path, old_category_id, new_name):
        """ Rename category. 
            @category_path contains path to the category node (last component
                           is renamed node).
            @old_category_id is id="" attribute of the renamed node
            @new_name is new title supplied by the user.
            Storage is expected to:
                - find old node by category_path and old_category_id
                - change category title
                - change category id to new generated id
                
            Returns new category id.
        """

class VirtualTreeCategoriesError(Exception): """ """


from Products.CMFCore.utils import getToolByName

def installHandler(self):
    if self.readDataFile('virtualtreecategories-install.txt') is None:
        return

    portal = self.getSite()
    qi = getToolByName(portal, 'portal_quickinstaller')
    if not qi.isProductInstalled('AddRemoveWidget'):
        # there is no GS profile for AddRemoveWidget.
        qi.installProduct('AddRemoveWidget')
    
def uninstall(self):
    if self.readDataFile('virtualtreecategories-uninstall.txt') is None:
        return

    portal = self.getSite()
    portal_conf=getToolByName(portal,'portal_controlpanel')
    portal_conf.unregisterConfiglet('VirtualTreeCategories')




from zope.interface import implements
from zope.component import adapts, getUtility
from zope.annotation import IAnnotations
from persistent.dict import PersistentDict
from persistent.list import PersistentList
from persistent import Persistent
from plone.i18n.normalizer.interfaces import IIDNormalizer
from Products.CMFPlone.utils import safe_unicode
from Products.CMFPlone.interfaces import IPloneSiteRoot
from collective.virtualtreecategories.interfaces import IVirtualTreeCategoryConfiguration, \
                                                        VirtualTreeCategoriesError
                                                        
from collective.virtualtreecategories.config import CATEGORY_SPLITTER, \
                                                    VTC_ANNOTATIONS_KEY
from BTrees.OOBTree import OOBTree
from zope.app.container.ordered import OrderedContainer

from types import ListType, TupleType
ListTypes = (ListType, TupleType)

import logging
logger = logging.getLogger('virtualtreecategories-storage')

class Category(OrderedContainer):
    id    = ''
    title = ''
    keywords = []
    
    def __init__(self, id, title=''):
        # XXX: This depends on implementation detail in OrderedContainer,
        # but it uses a PersistentDict, which sucks :-/
        OrderedContainer.__init__(self)
        self.id = id
        self.title = title or id
        self._data = OOBTree()

class VirtualTreeCategoryConfiguration(object):
    implements(IVirtualTreeCategoryConfiguration)
    adapts(IPloneSiteRoot)
    
    def __init__(self, context):
        self.context = context
        self.ann = IAnnotations(context)
        # category set here as root is not exposed to the public
        self.storage = self.ann.setdefault(VTC_ANNOTATIONS_KEY, Category('root-node', 'Root'))

    def _find_node(self, category_path):
        """ Returns node in path or root node """
        if not isinstance(category_path, ListTypes): 
            path = category_path.split(CATEGORY_SPLITTER) 
        else: 
            path = category_path 
        dpath = self.storage
        if category_path and path:
            # category_path may be empty string (root category)
            for item_id in path:
                dpath = dpath.get(item_id, None)
                if dpath is None:
                    return None
        return dpath

    def add_category(self, category_path, category_name):
        node = self._find_node(category_path)
        norm = getUtility(IIDNormalizer)
        category_id = norm.normalize(safe_unicode(category_name))
        if node.get(category_id, None) is not None:
            raise VirtualTreeCategoriesError('Category already exists')
        else:
            node[category_id] = Category(category_id, category_name)
            logger.info('Category %s (%s) added' % (category_name, category_id))
        return category_id

    def category_tree(self):
        def add_subkeys(node):
            res = []
            for k, category in node.items():
                item = dict(attributes = dict(id=category.id, rel='folder'),
                            state = 'closed',
                            data  = category.title,
                            children = add_subkeys(category)
                           )
                res.append(item)
            return res
        return add_subkeys(self.storage)
        
    def remove_category(self, category_path):
        node = self._find_node(category_path)
        if node is not None:
            parent = node.__parent__
            del parent[node.id]
            logger.info('Category %s (%s) removed' % (node.title, node.id))
            del node
            return True
        else:
            return False

    def rename_category(self, category_path, old_category_id, new_name):
        node = self._find_node(category_path)
        if node is not None:
            parent = node.__parent__
            norm = getUtility(IIDNormalizer)
            new_id = norm.normalize(safe_unicode(new_name))
            node.id = new_id
            node.title = new_name

            # Fix order of items (be sure item is on the same position as before)
            # Copy previous order but replace old_category_id with new id
            # Finally updateOrder of the parent's items
            new_order = []
            for item in list(parent.keys()):
                if item == old_category_id:
                    new_order.append(new_id)
                else:
                    new_order.append(item)
            del parent[old_category_id]
            parent[new_id] = node
            parent.updateOrder(new_order)
            logger.info('Category %s renamed to %s (%s)' % (old_category_id, new_name, new_id))
            return new_id
        else:
            return False
            
    def set(self, category_path, keywords):
        node = self._find_node(category_path)
        if node is not None:
            node.keywords = keywords
            logger.info('Keywords for category %s set to %r' % (node.title, keywords))
            return True
        else:
            return False

    def get(self, category_path):
        node = self._find_node(category_path)
        if node is not None:
            return node.keywords
        else:
            # this may happen eg. when new category is created
            return []



import types
from zope.component import ComponentLookupError
from Globals import InitializeClass
from AccessControl import ClassSecurityInfo
from Products.AddRemoveWidget.AddRemoveWidget import AddRemoveWidget
from Products.Archetypes.Registry import registerWidget

class VirtualTreeCategoriesWidget(AddRemoveWidget):
    _properties = AddRemoveWidget._properties.copy()
    _properties.update({
        'macro'                 : "virtualtreecategories_widget",
        'helper_js'             : ('widget_addremove_vars.js',
                                   'widget_addremove.js',
                                   "++resource++collective.virtualtreecategories.jsTree/css.js",
                                   "++resource++collective.virtualtreecategories.jsTree/tree_component.js",
                                   "virtualtreecategories_widget.js"),
        'helper_css'            : ('++resource++collective.virtualtreecategories.jsTree/tree_component.css',),
        })

    security = ClassSecurityInfo()

registerWidget(VirtualTreeCategoriesWidget,
               title='Virtual Tree Categories Widget',
               description=('You can filter out unwanted available keywords'),
               used_for=('Products.Archetypes.Field.LinesField',)
               )

# print "Rebinding ATDocument 'Subject' widget"
# from collective.virtualtreecategories.widget import VirtualTreeCategoriesWidget
# from Products.ATContentTypes.content.document import ATDocumentSchema
# old = ATDocumentSchema['subject'].widget
# ATDocumentSchema['subject'].widget = VirtualTreeCategoriesWidget(label=old.label, description=old.description)
# del old


from Products.PloneTestCase import ptc

from collective.testcaselayer import ptc as tcl_ptc

ptc.setupPloneSite()

class Layer(tcl_ptc.BasePTCLayer):
    """Install collective.virtualtreecategories"""

    def afterSetUp(self):
        self.addProfile('collective.virtualtreecategories:default')


from zope.annotation.attribute import AttributeAnnotations
from zope.component import provideAdapter
from collective.testcaselayer import ztc as tcl_ztc

class AttributeAnnotationsLayer(tcl_ztc.BaseZTCLayer):
    """Install annotations """

    def afterSetUp(self):
        provideAdapter(AttributeAnnotations)

layer = Layer([tcl_ptc.ptc_layer])
annotations_layer = AttributeAnnotationsLayer([tcl_ztc.ztc_layer])
